'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var compositionApi = require('@vue/composition-api');

function unwrap(o) {
    return compositionApi.isRef(o) ? o.value : o;
}
function wrap(o) {
    return compositionApi.isRef(o) ? o : compositionApi.ref(o);
}
const isArray = Array.isArray;
const isFunction = (val) => typeof val === "function";
const isString = (val) => typeof val === "string";
const isSymbol = (val) => typeof val === "symbol";
const isBoolean = (val) => typeof val === "boolean";
const isDate = (val) => isObject(val) && isFunction(val.getTime);
const isNumber = (val) => typeof val === "number";
const isObject = (val) => val !== null && typeof val === "object";
const isElement = (val) => isObject(val) && !!val.tagName;
function isPromise(val) {
    return isObject(val) && isFunction(val.then) && isFunction(val.catch);
}
const NO_OP = () => { };
const FALSE_OP = () => false;
const PASSIVE_EV = { passive: true };
function promisedTimeout(timeout) {
    return new Promise(res => {
        setTimeout(res, timeout);
    });
}
function minMax(val, min, max) {
    if (val < min)
        return min;
    if (val > max)
        return max;
    return val;
}
const isClient = typeof window != 'undefined';
// compact version: https://stackoverflow.com/a/33146982/1209882
/**
 * returns a random string
 * @param len length of the string max: 36
 */
// export function randomString(len: number) {
//   return (+new Date).toString(36).slice(-len);
// }

function usePagination(options) {
    const _currentPage = wrap(options.currentPage);
    const _pageSize = wrap(options.pageSize);
    const _offset = compositionApi.ref(0);
    const total = wrap(options.total);
    const offset = compositionApi.computed({
        get() {
            return _offset.value;
        },
        set(v) {
            if (!isNumber(v)) {
                return;
            }
            _offset.value = Math.min(v, total.value);
        }
    });
    const currentPage = compositionApi.computed({
        get() {
            return _currentPage.value;
        },
        set(v) {
            if (!isNumber(v)) {
                return;
            }
            _currentPage.value = minMax(v, 1, lastPage.value);
            // set the offset
            offset.value = (_currentPage.value - 1) * pageSize.value;
        }
    });
    const pageSize = compositionApi.computed({
        get() {
            return _pageSize.value;
        },
        set(v) {
            if (!isNumber(v)) {
                return;
            }
            _pageSize.value = v;
        }
    });
    const lastPage = compositionApi.computed(() => Math.ceil(total.value / pageSize.value));
    // make sure the current page is the correct value
    currentPage.value = _currentPage.value;
    const prev = () => --currentPage.value;
    const next = () => ++currentPage.value;
    const first = () => (currentPage.value = 1);
    const last = () => (currentPage.value = lastPage.value);
    compositionApi.watch([total, pageSize], () => {
        if (currentPage.value > lastPage.value) {
            currentPage.value = lastPage.value;
        }
    }, { lazy: true } // no need to run on first render
    );
    return {
        // Mutable state
        pageSize,
        total,
        currentPage,
        offset,
        // Computed
        lastPage,
        // Functions
        next,
        prev,
        first,
        last
    };
}

function useArrayPagination(array, options) {
    const arrayRef = wrap(array);
    const pagination = usePagination({
        ...{
            currentPage: 1,
            pageSize: 10,
        },
        ...options,
        total: compositionApi.computed(() => arrayRef.value.length)
    });
    const result = compositionApi.computed(() => {
        const array = arrayRef.value;
        if (!Array.isArray(array))
            return [];
        return array.slice(pagination.offset.value, pagination.offset.value + pagination.pageSize.value);
    });
    return {
        ...pagination,
        result
    };
}

function usePromise(fn, throwException = false) {
    if (!fn) {
        throw new Error(`[usePromise] argument can't be '${fn}'`);
    }
    if (typeof fn !== "function") {
        throw new Error(`[usePromise] expects function, but received ${typeof fn}`);
    }
    const loading = compositionApi.ref(false);
    const error = compositionApi.ref(null);
    const result = compositionApi.ref(null);
    const promise = compositionApi.ref();
    const exec = async (...args) => {
        loading.value = true;
        error.value = null;
        result.value = null;
        let throwExp = args && fn.length !== args.length && args.length > 0 && isBoolean(args[args.length - 1]) ? args[args.length - 1] : throwException;
        const currentPromise = (promise.value = fn(...args));
        try {
            const r = await currentPromise;
            if (promise.value === currentPromise) {
                result.value = r;
            }
            return r;
        }
        catch (er) {
            if (promise.value === currentPromise) {
                error.value = er;
                result.value = null;
            }
            return throwExp ? currentPromise : undefined;
        }
        finally {
            if (promise.value === currentPromise) {
                loading.value = false;
            }
        }
    };
    return {
        exec,
        result,
        promise,
        loading,
        error
    };
}

function useCancellablePromise(fn, throwException = false) {
    const cancelled = compositionApi.ref(false);
    let _cancel = undefined;
    const cancel = (result) => _cancel(result); // TODO add warning if cancel is undefined
    const promise = (p) => new Promise((res, rej) => {
        _cancel = result => {
            cancelled.value = true;
            rej(result);
        };
        p.then(res).catch(rej);
    });
    const use = usePromise((...args) => promise(fn(...args)), throwException);
    return {
        ...use,
        cancel,
        cancelled
    };
}

const MAX_RETRIES = 9000;
/* istanbul ignore next */
const ExecutionId = Symbol( ``);
/* istanbul ignore next */
const CancellationToken = Symbol( ``);
const defaultStrategy = async (options, context, factory, args) => {
    const retryId = context[ExecutionId].value;
    let i = -1;
    const maxRetries = options.maxRetries || MAX_RETRIES + 1;
    const delay = options.retryDelay || noDelay;
    context.retryErrors.value = [];
    context.isRetrying.value = false;
    context.nextRetry.value = undefined;
    let nextRetry = undefined;
    do {
        let success = false;
        let result = null;
        try {
            ++i;
            if (args) {
                result = factory(...args);
            }
            else {
                result = factory();
            }
            if (isPromise(result)) {
                result = await result;
            }
            // is cancelled?
            if (context[CancellationToken].value) {
                return null;
            }
            success = true;
        }
        catch (error) {
            result = null;
            context.retryErrors.value.push(error);
        }
        // is our retry current one?
        if (retryId !== context[ExecutionId].value) {
            return result;
        }
        if (success) {
            context.isRetrying.value = false;
            context.nextRetry.value = undefined;
            return result;
        }
        if (i >= maxRetries) {
            context.isRetrying.value = false;
            context.nextRetry.value = undefined;
            return Promise.reject(new Error(`[useRetry] max retries reached ${maxRetries}`));
        }
        context.isRetrying.value = true;
        const now = Date.now();
        const pDelayBy = delay(i); // wrapped
        const delayBy = isPromise(pDelayBy) ? await pDelayBy : pDelayBy; // unwrap promise
        if (!isPromise(pDelayBy) || !!delayBy) {
            if (isNumber(delayBy)) {
                nextRetry = delayBy;
            }
            else if (isDate(delayBy)) {
                nextRetry = delayBy.getTime();
            }
            else {
                throw new Error(`[useRetry] invalid value received from options.retryDelay '${typeof delayBy}'`);
            }
            // if the retry is in the past, means we need to await that amount
            if (nextRetry < now) {
                context.nextRetry.value = now + nextRetry;
            }
            else {
                context.nextRetry.value = nextRetry;
                nextRetry = nextRetry - now;
            }
            if (nextRetry > 0) {
                await promisedTimeout(nextRetry);
            }
        }
        // is cancelled?
        if (context[CancellationToken].value) {
            return null;
        }
        // is our retry current one?
        if (retryId !== context[ExecutionId].value) {
            return result;
        }
    } while (i < MAX_RETRIES);
    return null;
};
function useRetry(options, factory) {
    const opt = !options || isFunction(options) ? {} : options;
    const fn = isFunction(options) ? options : factory;
    if (!isFunction(options) && !isObject(options)) {
        throw new Error("[useRetry] options needs to be 'object'");
    }
    if (!!fn && !isFunction(fn)) {
        throw new Error("[useRetry] factory needs to be 'function'");
    }
    const isRetrying = compositionApi.ref(false);
    const nextRetry = compositionApi.ref();
    const retryErrors = compositionApi.ref([]);
    const cancellationToken = { value: false };
    const retryId = { value: 0 };
    const retryCount = compositionApi.computed(() => retryErrors.value.length);
    const context = {
        isRetrying,
        retryCount,
        nextRetry,
        retryErrors,
        [ExecutionId]: retryId,
        [CancellationToken]: cancellationToken
    };
    const exec = fn
        ? (...args) => {
            ++context[ExecutionId].value;
            return defaultStrategy(opt, context, fn, args);
        }
        : (f) => {
            ++context[ExecutionId].value;
            return defaultStrategy(opt, context, f, undefined);
        };
    const cancel = () => {
        context.isRetrying.value = false;
        context.retryErrors.value.push(new Error("[useRetry] cancelled"));
        context.nextRetry.value = undefined;
        cancellationToken.value = true;
    };
    return {
        ...context,
        cancel,
        exec
    };
}
const exponentialDelay = retryNumber => {
    const delay = Math.pow(2, retryNumber) * 100;
    const randomSum = delay * 0.2 * Math.random(); // 0-20% of the delay
    return delay + randomSum;
};
const noDelay = () => 0;

function useDebounce(handler, wait, options) {
    return debounce(handler, wait, options);
}
function debounce(func, waitMilliseconds = 50, options = {
    isImmediate: false
}) {
    let timeoutId;
    return function (...args) {
        const context = this;
        const doLater = function () {
            timeoutId = undefined;
            if (!options.isImmediate) {
                func.apply(context, args);
            }
        };
        const shouldCallNow = options.isImmediate && timeoutId === undefined;
        if (timeoutId !== undefined) {
            clearTimeout(timeoutId);
        }
        timeoutId = setTimeout(doLater, waitMilliseconds);
        if (shouldCallNow) {
            func.apply(context, args);
        }
    };
}

function useNow(options) {
    const SYNC_MS = 1000;
    const ms = options && options.refreshMs || SYNC_MS;
    const sync = options && isBoolean(options.sync) ? options.sync : true;
    const fn = options && isFunction(options.timeFn) && options.timeFn || Date.now;
    let handler = undefined;
    let timeoutHandler = undefined;
    const now = compositionApi.ref(fn());
    const remove = () => {
        clearInterval(handler);
        clearTimeout(timeoutHandler);
    };
    /* istanbul ignore next */
    const start = isClient ? () => handler = setInterval(() => now.value = fn(), ms) : NO_OP;
    if (sync) {
        const offset = SYNC_MS - (now.value - (Math.floor(now.value / SYNC_MS) * SYNC_MS));
        timeoutHandler = setTimeout(start, offset);
    }
    else {
        start();
    }
    compositionApi.onUnmounted(remove);
    return {
        now,
        remove
    };
}

function useDateNow(options) {
    const refreshMs = options && options.refreshMs || 1000;
    const sync = options && isBoolean(options.sync) ? options.sync : true;
    return useNow({
        refreshMs,
        sync,
        timeFn: Date.now
    });
}

function usePerformanceNow(options) {
    const refreshMs = options && options.refreshMs || 1000;
    const sync = options && isBoolean(options.sync) ? options.sync : true;
    return useNow({
        refreshMs,
        sync,
        timeFn: () => performance.now()
    });
}

exports.FALSE_OP = FALSE_OP;
exports.NO_OP = NO_OP;
exports.PASSIVE_EV = PASSIVE_EV;
exports.debounce = debounce;
exports.exponentialDelay = exponentialDelay;
exports.isArray = isArray;
exports.isBoolean = isBoolean;
exports.isClient = isClient;
exports.isDate = isDate;
exports.isElement = isElement;
exports.isFunction = isFunction;
exports.isNumber = isNumber;
exports.isObject = isObject;
exports.isPromise = isPromise;
exports.isString = isString;
exports.isSymbol = isSymbol;
exports.minMax = minMax;
exports.noDelay = noDelay;
exports.promisedTimeout = promisedTimeout;
exports.unwrap = unwrap;
exports.useArrayPagination = useArrayPagination;
exports.useCancellablePromise = useCancellablePromise;
exports.useDateNow = useDateNow;
exports.useDebounce = useDebounce;
exports.useNow = useNow;
exports.usePagination = usePagination;
exports.usePerformanceNow = usePerformanceNow;
exports.usePromise = usePromise;
exports.useRetry = useRetry;
exports.wrap = wrap;
