import { Ref } from '@vue/composition-api';

export declare interface ArrayPaginationResult<T extends Array<any>> extends PaginationResult {
    result: Readonly<Ref<T>>;
}

export declare interface CancellablePromiseResult<TCancel = any> {
    cancel: (result?: TCancel) => void;
    cancelled: Ref<boolean>;
}

declare const CancellationToken: unique symbol;

export declare function debounce<F extends Procedure>(func: F, waitMilliseconds?: number, options?: Options): F;

declare const ExecutionId: unique symbol;

export declare const exponentialDelay: RetryDelayFactory;

declare type Factory<T, TArgs extends Array<any>, TReturn = T> = (...args: TArgs) => TReturn;

export declare const FALSE_OP: () => boolean;

export declare const isArray: (arg: any) => arg is any[];

export declare const isBoolean: (val: unknown) => val is Boolean;

export declare const isClient: boolean;

export declare const isDate: (val: unknown) => val is Date;

export declare const isElement: (val: unknown) => val is Element;

export declare const isFunction: (val: unknown) => val is Function;

export declare const isNumber: (val: unknown) => val is number;

export declare const isObject: (val: unknown) => val is Record<any, any>;

export declare function isPromise<T = any>(val: unknown): val is Promise<T>;

export declare const isString: (val: unknown) => val is string;

export declare const isSymbol: (val: unknown) => val is symbol;

export declare function minMax(val: number, min: number, max: number): number;

export declare const NO_OP: () => void;

export declare const noDelay: RetryDelayFactory;

export declare interface NowOptions {
    /**
     * @description - interval in ms
     */
    refreshMs?: number;
    /**
     * @description - sync with seconds the clock
     */
    sync?: boolean;
}

export declare type Options = {
    isImmediate: boolean;
};

declare type PaginationControl = () => void;

export declare interface PaginationOptions {
    pageSize: RefTyped<number>;
    total: RefTyped<number>;
    currentPage: RefTyped<number>;
}

export declare interface PaginationResult {
    pageSize: Ref<number>;
    total: Ref<number>;
    currentPage: Ref<number>;
    offset: Ref<number>;
    lastPage: Readonly<Ref<number>>;
    next: PaginationControl;
    prev: PaginationControl;
    first: PaginationControl;
    last: PaginationControl;
}

export declare const PASSIVE_EV: AddEventListenerOptions;

/**
 * A function that emits a side effect and does not return anything.
 * https://github.com/chodorowicz/ts-debounce/blob/master/src/index.ts
 */
export declare type Procedure = (...args: any[]) => void;

export declare function promisedTimeout(timeout: number): Promise<void>;

declare interface PromiseResult<T extends Promise<any>, TR = PromiseType<T>, TError = any> {
    promise: Ref<T | undefined>;
    result: Ref<TR | null>;
    loading: Ref<boolean>;
    error: Ref<TError>;
}

export declare interface PromiseResultFactory<T extends Promise<any>, TArgs extends Array<any> = Array<any>> extends PromiseResult<T> {
    exec: (...args: TArgs) => Promise<PromiseType<T> | undefined>;
}

declare type PromiseType<T extends Promise<any>> = T extends Promise<infer R> ? R : never;

export declare type RefElement = Element | Ref<Element | undefined>;

export declare type RefTyped<T> = T | Ref<T>;

declare interface RetryContext {
    /**
     * @description Current attempt number
     */
    retryCount: Ref<number>;
    /**
     * @description When it should retry
     * @example new Date(nextRetry.value)
     */
    nextRetry: Ref<number | undefined>;
    /**
     * @description Get the current state, set to true after the first failure
     */
    isRetrying: Ref<boolean>;
    /**
     * @description List of all the errors occurred in since the last `exec` call
     */
    retryErrors: Ref<object[]>;
    /**
     * @description **INTERNAL** incremented every time `exec` is called
     */
    [ExecutionId]: {
        value: number;
    };
    /**
     * @description **INTERNAL** Used to cancel last retry
     */
    [CancellationToken]: {
        value: boolean;
    };
}

/**
 * @description How long should it delay or what time it should execute, it also supports Promises
 * @param {number} retry - Current retry number
 * @returns {number | Date | Promise<void> | Promise<number> | Promise<Date>} - Should wait until `Date`, wait until `Promise` finish or wait `number` ms
 */
declare type RetryDelayFactory = (retry: number) => number | Date | Promise<void> | Promise<number> | Promise<Date>;

declare interface RetryOptions {
    /**
     * @description Maximum of times it should retry
     */
    maxRetries?: number;
    /**
     * @description How long should it delay or what time it should execute, it also supports Promises
     */
    retryDelay?: RetryDelayFactory;
}

declare interface RetryReturn extends RetryContext {
    /**
     * @description Cancels last retry
     */
    cancel: () => void;
}

export declare interface RetryReturnFactory<T, TArgs extends Array<any>> extends RetryReturn {
    /**
     * Executes factory with the arguments passed
     * @param args - Arguments for the execution of the factory function passed at `useRetry`
     */
    exec(...args: TArgs): T;
}

export declare interface RetryReturnNoFactory extends RetryReturn {
    /**
     * Executes factory until success
     * @param fn - Factory function
     */
    exec<T>(fn: () => T): T;
}

export declare function unwrap(o: RefElement): Element;

export declare function unwrap<T>(o: RefTyped<T>): T;

export declare function useArrayPagination<T extends Array<TR>, TR>(array: RefTyped<T>, options?: Partial<Omit<PaginationOptions, 'total'>>): ArrayPaginationResult<T>;

export declare function useCancellablePromise<T extends any, TArgs extends Array<any>>(fn: (...args: TArgs) => Promise<T>): PromiseResultFactory<Promise<T>, TArgs> & CancellablePromiseResult;

export declare function useCancellablePromise<T extends any, TArgs extends Array<any>>(fn: (...args: TArgs) => Promise<T>, throwException: boolean): PromiseResultFactory<Promise<T>, TArgs> & CancellablePromiseResult;

export declare function useCancellablePromise<T extends any>(fn: () => T): PromiseResultFactory<Promise<T>> & CancellablePromiseResult;

export declare function useCancellablePromise<T extends any>(fn: () => T, throwException: boolean): PromiseResultFactory<Promise<T>> & CancellablePromiseResult;

export declare function useCancellablePromise<T extends Promise<TR>, TR, TArgs extends Array<any>>(fn: (...args: TArgs) => T): PromiseResultFactory<T, TArgs> & CancellablePromiseResult;

export declare function useCancellablePromise<T extends Promise<TR>, TR, TArgs extends Array<any>>(fn: (...args: TArgs) => T, throwException: boolean): PromiseResultFactory<T, TArgs> & CancellablePromiseResult;

export declare function useCancellablePromise<T = any>(fn: () => T): PromiseResultFactory<Promise<T>> & CancellablePromiseResult;

export declare function useCancellablePromise<T = any>(fn: () => T, throwException: boolean): PromiseResultFactory<Promise<T>> & CancellablePromiseResult;

export declare function useCancellablePromise<T extends Promise<TR>, TR>(fn: () => T): PromiseResultFactory<T> & CancellablePromiseResult;

export declare function useDateNow(options?: NowOptions): {
    now: import("@vue/composition-api").Ref<number>;
    remove: () => void;
};

export declare function useDebounce<T extends Function>(handler: T, wait?: number, options?: Options): T;

export declare function useNow(options?: NowOptions & UseNowOptions): {
    now: import("@vue/composition-api").Ref<number>;
    remove: () => void;
};

export declare interface UseNowOptions {
    /**
     * @description - time factory, it should retrieve `now` in ms
     */
    timeFn?: () => number;
}

export declare function usePagination(options: PaginationOptions): PaginationResult;

export declare function usePerformanceNow(options?: NowOptions): {
    now: import("@vue/composition-api").Ref<number>;
    remove: () => void;
};

/**
 *
 * @param fn - factory function
 * @param throwException - if `true` allows to catch exception when `exec()`
 */
export declare function usePromise<T = any, TArgs extends Array<any> = Array<any>>(fn: (...args: TArgs) => Promise<T>, throwException?: boolean): PromiseResultFactory<Promise<T>, TArgs>;

export declare function usePromise<T = any, TArgs extends Array<any> = Array<any>>(fn: (...args: TArgs) => Promise<T>): PromiseResultFactory<Promise<T>, TArgs>;

export declare function usePromise<T = any, TArgs extends Array<any> = Array<any>>(fn: (...args: TArgs) => T, throwException: boolean): PromiseResultFactory<Promise<T>, TArgs>;

export declare function usePromise<T = any, TArgs extends Array<any> = Array<any>>(fn: (...args: TArgs) => T): PromiseResultFactory<Promise<T>, TArgs>;

export declare function usePromise<T = any>(fn: () => Promise<T>, throwException: boolean): PromiseResultFactory<Promise<T>>;

export declare function usePromise<T = any>(fn: () => Promise<T>): PromiseResultFactory<Promise<T>>;

export declare function usePromise<T = any>(fn: () => T, throwException: boolean): PromiseResultFactory<Promise<T>>;

export declare function usePromise<T = any>(fn: () => T): PromiseResultFactory<Promise<T>>;

export declare function useRetry(options?: RetryOptions): RetryReturnNoFactory;

export declare function useRetry<T, TArgs extends Array<any>>(factory: Factory<T, TArgs>): RetryReturnFactory<T, TArgs>;

export declare function useRetry<T, TArgs extends Array<any>>(options: RetryOptions, factory: Factory<T, TArgs>): RetryReturnFactory<T, TArgs>;

export declare function wrap(o: RefElement): Ref<Element>;

export declare function wrap<T>(o: RefTyped<T>): Ref<T>;

export { }
